type arbre = Feuille of int * int
           | Noeud of int * int * arbre * arbre ;;

(*
   Paramètre = arbre : un arbre binaire
   Renvoie un triplé ( n, f, h ) où
    - n est le nombre de noeud de l'arbre
    - f est le nombre de feuilles de l'arbre
    - h est la hauteur de l'arbre
*)
let rec grandeurs_arbre ( arbre : arbre ) : int * int * int =
  match arbre with
  | Feuille( _, _ ) -> ( 1, 1, 1 )
  | Noeud ( _, _, fils_gauche, fils_droit ) ->
        let nb_noeuds_gauche, nb_feuilles_gauche, hauteur_gauche =
          ( grandeurs_arbre fils_gauche ) in
        let nb_noeuds_droit, nb_feuilles_droit, hauteur_droit =
          ( grandeurs_arbre fils_droit ) in
        ( 1 + nb_noeuds_gauche + nb_noeuds_droit, (* Nombre de noeuds *)
          nb_feuilles_gauche + nb_feuilles_droit, (* Nombre de feuilles *)
          1 + ( max hauteur_gauche hauteur_droit ) ) (* Hauteur de l'arbre *)
;;

(*
Paramètre = arbre : un arbre binaire entier non vide
Renvoie l'adresse de la racine de l'arbre
*)
let adresse_racine ( arbre : arbre ) : int * int = match arbre with
  | Feuille ( l, m ) -> ( l, m )
  | Noeud ( l, m, _, _ ) -> ( l, m ) ;;

(*
Paramètre = adresse : l'adresse du noeud à chercher
Paramètre = arbre : l'arbre dans lequel chercher l'adresse
Renvoie l'étiquette du noeud à l'adresse passée en paramètre'
*)
let rec chercher_noeud ( arbre : arbre ) ( adresse : int ) : int * int =
  match arbre with
  | Feuille ( l, m ) when l != adresse ->
      raise ( Invalid_argument "Adresse non présente dans l'arbre" )
  | Feuille ( l, m ) -> ( l, m )
  | Noeud ( l, m, _, _ ) when l == adresse -> ( l, m )
  | Noeud ( _, _, fils_g, fils_d ) ->
      let _, m_g  = adresse_racine fils_g in
      if adresse <= m_g then chercher_noeud fils_g adresse
      else chercher_noeud fils_d adresse ;;

(*
   Paramètre = arbre : l'arbre dans lequel le calcul de distance a lieu
   Paramètre = adresse : l'adresse du noeud dont on calcul la distance
   Renvoie la distance à la racine de l'arbre du noeud d'adresse <adresse>
*)
let distance_a_la_racine ( arbre : arbre ) ( adresse : int ) : int =
    (*
        Paramètre = arbre : l'arbre dans lequel le calcul de distance a lieu
        Paramètre = distance : la distance accumulée depuis le début du parcours
        Renvoie <distance> + la distance à la racine de l'arbre du noeud
            d'adresse <adresse>
    *)
  let rec recherche ( arbre : arbre ) ( distance : int ) : int =
    match arbre with
    | Feuille ( l, m ) when l != adresse ->
        raise ( Invalid_argument "Adresse non présente dans l'arbre" )
    | Feuille ( l, m ) -> distance
    | Noeud ( l, m, _, _ ) when l == adresse -> distance
    | Noeud ( _, _, fils_g, fils_d ) ->
        let _, m_g = adresse_racine fils_g in
        if adresse <= m_g then recherche fils_g ( distance + 1 )
        else recherche fils_d ( distance + 1 ) in
  recherche arbre 0 ;;

(*
   Paramètre = arbre : l'arbre dans lequel le calcul de distance a lieu
   Paramètre = adresse_source : l'adresse du noeud de départ
   Paramètre = adresse_destination : l'adresse du noeud de destination
   Renvoie la distance du noeud de départ au noeud de destination dans l'arbre
*)
let rec distance
    ( arbre : arbre  )
    ( adresse_source : int )
    ( adresse_destination : int )
    : int =
  match arbre with
    | Feuille ( l, m ) when l != adresse_source || l != adresse_destination ->
        raise ( Invalid_argument "Incohérence dans l'arbre" )
    | Feuille ( l, m ) -> 0
    | Noeud ( l, m, fils_g, fils_d ) ->
        let l_g, m_g = adresse_racine fils_g in
        if ( l_g <= adresse_source && adresse_source <= m_g ) &&
           ( l_g <= adresse_destination && adresse_destination <= m_g )
        then distance fils_g adresse_source adresse_destination
        else let l_d, m_d = adresse_racine fils_d in
            if ( l_d <= adresse_source && adresse_source <= m_d ) &&
            ( l_d <= adresse_destination && adresse_destination <= m_d )
            then distance fils_d adresse_source adresse_destination
            else ( distance_a_la_racine arbre adresse_source )
                 + ( distance_a_la_racine arbre adresse_destination )
                 - 1 (* la racine est parcouru deux fois sinon *) ;;
